@@ -1,5 +1,6 @@
 import asyncio
 import json
+import logging
 from inspect import isawaitable
 from typing import (
     Any,
@@ -9,18 +10,20 @@ from typing import (
     Dict,
     List,
     Optional,
+    Sequence,
+    Type,
     Union,
     cast,
 )
 
 import graphene
 from graphql import (
+    ExecutionContext,
     ExecutionResult,
     GraphQLError,
     Middleware,
     OperationType,
     execute,
-    format_error,
     graphql,
     parse,
     subscribe,
@@ -35,6 +38,15 @@ from starlette.responses import HTMLResponse, JSONResponse, Response
 from starlette.types import Receive, Scope, Send
 from starlette.websockets import WebSocket, WebSocketDisconnect, WebSocketState
 
+try:
+    # graphql-core==3.2.*
+    from graphql import GraphQLFormattedError
+    from graphql.error.graphql_error import format_error
+except ImportError:
+    # graphql-core==3.1.*
+    from graphql import format_error
+    GraphQLFormattedError = Dict[str, Any]
+
 GQL_CONNECTION_ACK = "connection_ack"
 GQL_CONNECTION_ERROR = "connection_error"
 GQL_CONNECTION_INIT = "connection_init"
@@ -49,14 +61,16 @@ ContextValue = Union[Any, Callable[[HTTPConnection], Any]]
 RootValue = Any
 
 
-def make_graphiql_handler():
+def make_graphiql_handler() -> Callable[[Request], Response]:
     def handler(request: Request) -> Response:
         return HTMLResponse(_GRAPHIQL_HTML)
 
     return handler
 
 
-def make_playground_handler(playground_options=None):
+def make_playground_handler(
+    playground_options: Optional[Dict[str, Any]] = None
+) -> Callable[[Request], Response]:
     playground_options_str = json.dumps(playground_options or {})
     content = _PLAYGROUND_HTML.replace("PLAYGROUND_OPTIONS", playground_options_str)
 
@@ -70,19 +84,26 @@ class GraphQLApp:
     def __init__(
         self,
         schema: graphene.Schema,
-        on_get: Union[
-            None, Callable[[Request], Union[Response, Awaitable[Response]]]
+        *,
+        on_get: Optional[
+            Callable[[Request], Union[Response, Awaitable[Response]]]
         ] = None,
         context_value: ContextValue = None,
         root_value: RootValue = None,
         middleware: Optional[Middleware] = None,
-        playground: bool = False,  # deprecating
+        error_formatter: Callable[[GraphQLError], GraphQLFormattedError] = format_error,
+        logger_name: Optional[str] = None,
+        execution_context_class: Optional[Type[ExecutionContext]] = None,
+        playground: bool = False,  # Deprecating. Use on_get instead.
     ):
         self.schema = schema
         self.on_get = on_get
         self.context_value = context_value
         self.root_value = root_value
+        self.error_formatter = error_formatter
         self.middleware = middleware
+        self.execution_context_class = execution_context_class
+        self.logger = logging.getLogger(logger_name or __name__)
 
         if playground and self.on_get is None:
             self.on_get = make_playground_handler()
@@ -115,7 +136,7 @@ class GraphQLApp:
 
         response = handler(request)
         if isawaitable(response):
-            return await cast(Awaitable, response)
+            return await cast(Awaitable[Response], response)
         else:
             return cast(Response, response)
 
@@ -134,8 +155,8 @@ class GraphQLApp:
     async def _handle_http_request(self, request: Request) -> JSONResponse:
         try:
             operations = await _get_operation_from_request(request)
-        except ValueError as error:
-            return JSONResponse({"errors": [error.args[0]]}, status_code=400)
+        except ValueError as e:
+            return JSONResponse({"errors": [e.args[0]]}, status_code=400)
 
         if isinstance(operations, list):
             return JSONResponse(
@@ -157,11 +178,20 @@ class GraphQLApp:
             middleware=self.middleware,
             variable_values=variable_values,
             operation_name=operation_name,
+            execution_context_class=self.execution_context_class,
         )
 
         response: Dict[str, Any] = {"data": result.data}
         if result.errors:
-            response["errors"] = [format_error(error) for error in result.errors]
+            for error in result.errors:
+                if error.original_error:
+                    self.logger.error(
+                        "An exception occurred in resolvers",
+                        exc_info=error.original_error,
+                    )
+            response["errors"] = [
+                self.error_formatter(error) for error in result.errors
+            ]
 
         return JSONResponse(
             response,
@@ -170,7 +200,7 @@ class GraphQLApp:
         )
 
     async def _run_websocket_server(self, websocket: WebSocket) -> None:
-        subscriptions: Dict[str, AsyncGenerator] = {}
+        subscriptions: Dict[str, AsyncGenerator[Any, None]] = {}
         await websocket.accept("graphql-ws")
         try:
             while (
@@ -192,10 +222,10 @@ class GraphQLApp:
 
     async def _handle_websocket_message(
         self,
-        message: dict,
+        message: Dict[str, Any],
         websocket: WebSocket,
-        subscriptions: Dict[str, AsyncGenerator],
-    ):
+        subscriptions: Dict[str, AsyncGenerator[Any, None]],
+    ) -> None:
         operation_id = cast(str, message.get("id"))
         message_type = cast(str, message.get("type"))
 
@@ -218,8 +248,8 @@ class GraphQLApp:
         data: Any,
         operation_id: str,
         websocket: WebSocket,
-        subscriptions: Dict[str, AsyncGenerator],
-    ):
+        subscriptions: Dict[str, AsyncGenerator[Any, None]],
+    ) -> None:
         query = data["query"]
         variable_values = data.get("variables")
         operation_name = data.get("operationName")
@@ -236,6 +266,7 @@ class GraphQLApp:
             errors = [e]
 
         if not errors:
+            assert document is not None
             if operation and operation.operation == OperationType.SUBSCRIPTION:
                 errors = await self._start_subscription(
                     websocket,
@@ -250,7 +281,6 @@ class GraphQLApp:
                 errors = await self._handle_query_over_ws(
                     websocket,
                     operation_id,
-                    subscriptions,
                     document,
                     context_value,
                     variable_values,
@@ -262,19 +292,18 @@ class GraphQLApp:
                 {
                     "type": GQL_ERROR,
                     "id": operation_id,
-                    "payload": format_error(errors[0]),
+                    "payload": self.error_formatter(errors[0]),
                 }
             )
 
     async def _handle_query_over_ws(
         self,
-        websocket,
-        operation_id,
-        subscriptions,
-        document,
-        context_value,
-        variable_values,
-        operation_name,
+        websocket: WebSocket,
+        operation_id: str,
+        document: DocumentNode,
+        context_value: ContextValue,
+        variable_values: Dict[str, Any],
+        operation_name: str,
     ) -> List[GraphQLError]:
         result = execute(
             self.schema.graphql_schema,
@@ -284,6 +313,7 @@ class GraphQLApp:
             variable_values=variable_values,
             operation_name=operation_name,
             middleware=self.middleware,
+            execution_context_class=self.execution_context_class,
         )
 
         if isinstance(result, ExecutionResult) and result.errors:
@@ -297,7 +327,13 @@ class GraphQLApp:
         payload: Dict[str, Any] = {}
         payload["data"] = result.data
         if result.errors:
-            payload["errors"] = [format_error(error) for error in result.errors]
+            for error in result.errors:
+                if error.original_error:
+                    self.logger.error(
+                        "An exception occurred in resolvers",
+                        exc_info=error.original_error,
+                    )
+            payload["errors"] = [self.error_formatter(error) for error in result.errors]
 
         await websocket.send_json(
             {"type": GQL_DATA, "id": operation_id, "payload": payload}
@@ -306,13 +342,13 @@ class GraphQLApp:
 
     async def _start_subscription(
         self,
-        websocket,
-        operation_id,
-        subscriptions,
-        document,
-        context_value,
-        variable_values,
-        operation_name,
+        websocket: WebSocket,
+        operation_id: str,
+        subscriptions: Dict[str, AsyncGenerator[Any, None]],
+        document: DocumentNode,
+        context_value: ContextValue,
+        variable_values: Dict[str, Any],
+        operation_name: str,
     ) -> List[GraphQLError]:
         result = await subscribe(
             self.schema.graphql_schema,
@@ -326,7 +362,7 @@ class GraphQLApp:
         if isinstance(result, ExecutionResult) and result.errors:
             return result.errors
 
-        asyncgen = cast(AsyncGenerator, result)
+        asyncgen = cast(AsyncGenerator[Any, None], result)
         subscriptions[operation_id] = asyncgen
         asyncio.create_task(
             self._observe_subscription(asyncgen, operation_id, websocket)
@@ -334,23 +370,26 @@ class GraphQLApp:
         return []
 
     async def _observe_subscription(
-        self, asyncgen: AsyncGenerator, operation_id: str, websocket: WebSocket
+        self,
+        asyncgen: AsyncGenerator[Any, None],
+        operation_id: str,
+        websocket: WebSocket,
     ) -> None:
         try:
             async for result in asyncgen:
-                payload = {}
-                payload["data"] = result.data
+                payload = {"data": result.data}
                 await websocket.send_json(
                     {"type": GQL_DATA, "id": operation_id, "payload": payload}
                 )
         except Exception as error:
             if not isinstance(error, GraphQLError):
+                self.logger.error("An exception occurred in resolvers", exc_info=error)
                 error = GraphQLError(str(error), original_error=error)
             await websocket.send_json(
                 {
                     "type": GQL_DATA,
                     "id": operation_id,
-                    "payload": {"errors": [format_error(error)]},
+                    "payload": {"errors": [self.error_formatter(error)]},
                 }
             )
 
@@ -361,11 +400,13 @@ class GraphQLApp:
             await websocket.send_json({"type": GQL_COMPLETE, "id": operation_id})
 
 
-async def _get_operation_from_request(request: Request):
+async def _get_operation_from_request(
+    request: Request,
+) -> Union[Dict[str, Any], List[Any]]:
     content_type = request.headers.get("Content-Type", "").split(";")[0]
     if content_type == "application/json":
         try:
-            return await request.json()
+            return cast(Union[Dict[str, Any], List[Any]], await request.json())
         except (TypeError, ValueError):
             raise ValueError("Request body is not a valid JSON")
     elif content_type == "multipart/form-data":
@@ -374,7 +415,9 @@ async def _get_operation_from_request(request: Request):
         raise ValueError("Content-type must be application/json or multipart/form-data")
 
 
-async def _get_operation_from_multipart(request: Request):
+async def _get_operation_from_multipart(
+    request: Request,
+) -> Union[Dict[str, Any], List[Any]]:
     try:
         request_body = await request.form()
     except Exception:
@@ -396,19 +439,28 @@ async def _get_operation_from_multipart(request: Request):
 
     files = {k: v for (k, v) in request_body.items() if isinstance(v, UploadFile)}
     for (name, paths) in name_path_map.items():
+        file = files.get(name)
+        if not file:
+            raise ValueError(
+                f"File fields don't contain a valid UploadFile type for '{name}' mapping"
+            )
+
         for path in paths:
             path = tuple(path.split("."))
-            _inject_file_to_operations(operations, files[name], path)
+            _inject_file_to_operations(operations, file, path)
 
     return operations
 
 
-def _inject_file_to_operations(ops_tree, _file, path):
-    key = path[0]
+def _inject_file_to_operations(
+    ops_tree: Any, _file: UploadFile, path: Sequence[str]
+) -> None:
+    k = path[0]
+    key: Union[str, int]
     try:
-        key = int(key)
+        key = int(k)
     except ValueError:
-        pass
+        key = k
     if len(path) == 1:
         if ops_tree[key] is None:
             ops_tree[key] = _file
